@@ -25,7 +25,7 @@ use Config::Objective::DataType;
 use Config::Objective::Parser;
 
 
-our $VERSION = '0.9.1';
+our $VERSION = '0.5';
 our $AUTOLOAD;
 
 
@@ -80,11 +80,6 @@ sub _call_obj_method
 	my ($self, $obj, $method, @args) = @_;
 	my ($retval, $line, $msg);
 
-#	print "==> _call_obj_method('$obj', '$method'";
-#	map { print ", '$_'"; } @args
-#		if (@args > 1 || defined($args[0]));
-#	print ")\n";
-
 	die "$obj: unknown config object"
 		if (!exists($self->{'objs'}->{$obj}));
 
@@ -103,8 +98,6 @@ sub _call_obj_method
 		die "$msg: $@";
 	}
 
-#	print "<== _call_obj_method(): returning '"
-#		. (defined($retval) ? $retval : 'undef') . "'\n";
 	return $retval;
 }
 
@@ -133,8 +126,6 @@ sub new
 	$self->{'list_stack'} = [];
 	$self->{'hash_stack'} = [];
 
-	$self->{'in_expr'} = 0;
-
 	$self->parse($file);
 
 	return $self;
@@ -159,13 +150,13 @@ sub parse
 	$lexer = Parse::Lex->new(
 		'AND',		'&&',
 		'COMMA',	',',
-		'COMMENT',	'(?<!\\\\)#.*$',
+		'COMMENT',	'(?<!\\$)#.*$',
 		'ELIF',		'^\%[ \t]*elif',
 		'ELSE',		'^\%[ \t]*else',
 		'ENDIF',	'^\%[ \t]*endif',
 		'EOS',		';',
-		'PAREN_START',	'\(',
-		'PAREN_END',	'\)',
+		'EXPR_START',	'\(',
+		'EXPR_END',	'\)',
 		'HASH_ARROW',	'=>',
 		'HASH_START',	'{',
 		'HASH_END',	'}',
@@ -174,30 +165,22 @@ sub parse
 		'LIST_START',	'\[',
 		'LIST_END',	'\]',
 		'METHOD_ARROW',	'->',
-		'NOT',		'!',
 		'OR',		'\|\|',
 		'WORD',		'\w+',
-		'QSTRING',	[
-				  '(?<!\\\\)"',
-				  '([^"]|(?<=\\\\)")*',
-				  '(?<!\\\\)"'
-				],
+		'QSTRING',	[ '"', '[^"]*', '"' ],
 				sub {
 					my ($token, $string) = @_;
 
-					$string =~ s/^"//;
-					$string =~ s/"$//;
-
-					$string =~ s/\\"/"/g;
+					$string =~ s/^\"//;
+					$string =~ s/\"$//;
 
 					return $string;
 				},
 		'ERROR',	'(?s:.*)',
 				sub {
-					my $line = $_[0]->lexer->line;
-					my $file = $self->{'file_stack'}->[-1];
+					my $line = $_[0]->line;
 
-					die "$file:$line: syntax error: \"$_[1]\"\n";
+					die "line $line: syntax error: \"$_[1]\"\n";
 				}
 	);
 	$lexer->from(\*$fh);
@@ -300,12 +283,10 @@ Config::Objective - Perl module for parsing object-oriented config files
 =head1 SYNOPSIS
 
   use Config::Objective;
-  use Config::Objective::String;
-  use Config::Objective::List;
 
   my $conf = Config::Objective->new('filename',
   		{
-			'var1' => Config::Objective::String->new(),
+			'var1' => Config::Objective::Scalar->new(),
 			'var2' => Config::Objective::List->new(),
 			...
 		},
@@ -364,78 +345,12 @@ are the result of calling the get() method on the corresponding object.
 The config file format supported by B<Config::Objective> is described
 here.
 
-=head2 Data Types
-
-B<Config::Objective> supports three types of data: scalars, lists, and
-hashes.  The syntax for these types is intentionally similar to their
-perl equivalents.
-
-=over 4
-
-=item Scalars
-
-A scalar is represented as a simple integer or string value.  If it is
-composed only of letters, numbers, and the underscore character, it can
-be written literally:
-
-  foo
-  all_word_characters
-  123
-  alpha_123_numeric
-  4sure
-
-However, if the value contains whitespace or other non-word characters,
-it must be quoted:
-
-  "telnet/tcp"
-  "use quotes for whitespace"
-  "quotes can be escaped like this \" inside quoted strings"
-  "quoted
-     strings
-   can span
-   multiple lines"
-
-=item Lists
-
-A list is represented as a sequence of comma-delimited values enclosed
-by square brackets:
-
-  [ this, is, a, list ]
-
-Note that each value in a list can itself be a scalar, list, or hash:
-
-  [ this, is, a, [ nested, list ] ]
-  [ this, list, contains, a, { hash => value } ]
-
-=item Hashes
-
-A hash is represented as a sequence of zero or more comma-delimited
-entries enclosed in curly braces:
-
-  { this => 1, is => 2, a => 3, hash => 4 }
-
-As in perl, each entry contains a key and a value.  However, unlike perl,
-the value is optional:
-
-  { this, is, a, hash, without, values }
-  { this => hash, has => "some values", but, not, others }
-
-When no value is specified for a given entry, its value is undefined.
-
-Note that hash keys must always be scalars.  However, values may be
-scalars, lists, or hashes:
-
-  { "this is a" => [ list, within, a, hash ] }
-  { "this is a" => { sub => hash } }
-
-=back
-
 =head2 Configuration Statements
 
 Each statement in the config file results in calling a method on a
 configuration object.  The syntax is:
 
-  object[->method] [args];
+  object[->method] [arg];
 
 In this syntax, "object" is the name of the object.  The object must
 be created and passed to the B<Config::Objective> constructor, as
@@ -445,35 +360,27 @@ The "->method" portion is optional.  If specified, it indicates which
 method should be called on the object.  If not specified, a method called
 default() will be used.
 
-The "args" portion is also optional.  It specifies one or more
-comma-delimited arguments to pass to the method.  If multiple arguments
-are provided, the entire argument list must be enclosed in parentheses.
-Each argument can be a simple scalar, list, hash, or a complex, nested
-list or hash structure, as described above.
+The "arg" portion is also optional.  It specifies an argument to pass to
+the method.  It can be a simple scalar, list, hash, or a complex, nested
+list or hash structure.  For example:
+
+  all_word_characters
+  "use quotes for non-word characters"
+  [ this, is, a, list ]
+  { this => 1, is => 2, a => 3, hash => 4 }
+  { hash, values => are, optional }
+  [ this, is, a, [ nested, list ] ]
+  [ this, is, a, { hash, within, a, list } ]
+  { "this is a" => [ list, within, a, hash ] }
 
 So, putting this all together, here are some example configuration
 statements:
 
-  ### call default method with no arguments
-  object;
-
-  ### call a specific method, but still no args
-  object->method;
-
-  ### call default method, but specify a single scalar argument
-  object scalar_arg;
-
-  ### call default method, but specify a single list argument
-  object [ this, is, a, single, list, argument ];
-
-  ### call a specific method and specify a single hash argument
-  object->method { this, is, a, single, hash, argument };
-
-  ### call a specific method with multiple scalar args
-  object->method(arg1, arg2, arg3);
-
-  ### call a specific method with multiple args of different types
-  object->method(scalar_arg, [ list, argument ], { hash => argument });
+  object_default_method_no_args;
+  object_no_args->some_method;
+  object_default_method scalar_arg;
+  object [ list, for, args ];
+  object_method_and_args->another_method { hash, for, args };
 
 =head2 Conditional Evaluation
 
@@ -489,24 +396,23 @@ They can be used to enclose other config statements, which are evaluated
 or skipped based on whether the conditional expression evaluates to true.
 For example:
 
-  %if ( expression )
+  %if expression
     ... other config directives ...
   %endif
 
 The most basic I<expression> is simply a method call that returns
 a true or false value.  The syntax for this is the same as a normal
-config statement, except without the trailing semicolon:
+config statement, except without the trailing semicolon.  For example:
 
-  %if ( object[->method] [args] )
+  %if object[->method] [arg]
 
 If no method is specified, the equals() method will be called by
 default.
 
-Multiple expressions can be combined using the "&&", "||", and
-"!" operators.  Additional parentheses can also be used for grouping
-within the expression.  For example:
+Multiple expressions can be combined using the "&&" and "||" operators.
+Parentheses can also be used for grouping.  For example:
 
-  %if ( ( object1 foo && ! object2 bar ) || object3 baz )
+  %if ( object1 foo && object2 bar ) || object3 baz
 
 =head2 File Inclusion
 
@@ -521,11 +427,11 @@ If the specified filename is not an absolute path, B<Config::Objective>
 will look for it in the directory specified by the I<include_dir>
 attribute when the B<Config::Objective> object was created.
 
-Note that the I<%include> directive will be ignored within an I<%if>
-block whose condition is false.  This means that you cannot start an
-I<%if> block in one file, add a I<%include> directive, and provide the
-I<%endif> directive in the included file.  All I<%if> blocks must be
-fully contained within the same file.
+Note that the "%include" directive will be ignored within an "%if" block
+whose condition is false.  This means that you cannot start an "%if"
+block in one file, add a "%include" directive, and provide the "%endif"
+directive in the included file.  All "%if" blocks must be fully
+contained within the same file.
 
 =head2 Comments
 
@@ -574,25 +480,211 @@ Evaluation"> above.)
 =head2 Supplied Object Classes
 
 B<Config::Objective> supplies several classes that can be used for
-encapsulating common types of configuration data.
+encapsulating common configuration data.
 
 =over 4
 
-=item B<Config::Objective::Boolean>
+=item B<Config::Objective::DataType>
 
-=item B<Config::Objective::Hash>
+This is the base class for the rest of the supplied config object classes.
+It should not be used directly, but does support the following methods
+for use in subclasses:
 
-=item B<Config::Objective::Integer>
+=over 4
 
-=item B<Config::Objective::List>
+=item new()
 
-=item B<Config::Objective::String>
+The constructor.  It can be passed a hash to set the object's
+attributes.  The object will be created as a reference to this hash.
 
-=item B<Config::Objective::Table>
+=item get()
+
+Returns the value encapsulated by the object.
+
+=item set()
+
+Sets the value encapsulated by the object to its argument.
+
+=item default()
+
+Calls the set() method.
+
+=item equals()
+
+Returns true if the argument equals the object's value (as determined by
+the I<eq> operator; see the L<perlop> man page for details).
+
+=item unset()
+
+Sets the object's value to I<undef>.
 
 =back
 
-See the documentation for each of these classes for more information.
+=item B<Config::Objective::Scalar>
+
+This object encapsulates a scalar value.  It supports the following
+methods:
+
+=over 4
+
+=item set()
+
+Sets the object's value to its argument.  The value must be a scalar.
+
+If the object was created with the I<value_abspath> attribute enabled,
+the value must be an absolute path string.
+
+If the object was created with the I<value_optional> attribute enabled,
+the argument is optional; if missing, an empty string will be used
+instead.
+
+=item append()
+
+Appends its argument to the object's value using string concatenation.
+
+=back
+
+=item B<Config::Objective::Boolean>
+
+This object encapsulates a boolean value.  It supports the following
+methods:
+
+=over 4
+
+=item set()
+
+Sets the object's value to its argument.  The value must be one of the
+following: "yes", "no", "on", "off", "true", "false", 1, or 0.
+
+=back
+
+=item B<Config::Objective::List>
+
+This object encapsulates a list value.  It supports the following
+methods:
+
+=over 4
+
+=item unset()
+
+Sets the object's value to a reference to an empty list.
+
+=item set()
+
+Sets the object's value to its argument, which must be a reference to
+a list.
+
+=item add()
+
+Appends its argument to the list.  The argument can be a scalar or a
+reference to a list, in which case the referenced list's content is
+added to the object's list.
+
+=item default()
+
+Calls the add() method.
+
+=item add_top()
+
+Same as add(), but adds to the front of the list instead of the end.
+
+=item delete()
+
+Deletes elements from the list that match its argument.  Matching is
+performed by using the argument as a regular expression.  The argument
+can be a scalar or a reference to a list, in which case each item of the
+referenced list is used to check the values in the object's list.
+
+=back
+
+=item B<Config::Objective::Table>
+
+This object encapsulates a table, which is represented as a list of lists.
+Both rows and columns are indexed starting at 0.  It is derived from
+the B<Config::Objective::List> class, but it supports the following
+additional methods:
+
+=over 4
+
+=item add_before()
+
+Inserts a new row into the table before a specified row.  The argument
+must be a reference to a list containing three elements: a number
+indicating what column to search on, a string which is used as a regular
+expression match to find a matching row in the table, and a reference
+to the new list to be inserted before the matching row.
+
+=item find()
+
+Finds a row with a specified word in a specified column.  The column
+number is the first argument, and the word to match on is the second.
+It returns a reference to the matching row, or I<undef> if no matches
+were found.
+
+This function is not very useful for calling from a config file, but
+it's sometimes useful to call it from perl once the config file has been
+read.
+
+=item replace()
+
+Finds a row in the same manner as find(), and then replaces that row's
+value in a specified column with a new value.  The arguments are the
+column number to search on, the word to search for, the column number to
+replace, and the text to replace it with.
+
+=item modify()
+
+Similar to replace(), but appends to the existing value instead of
+replacing it.  A space character is appended before the new value.
+
+=back
+
+=item B<Config::Objective::Hash>
+
+This object encapsulates a hash.  It supports the following methods:
+
+=over 4
+
+=item insert()
+
+Inserts the specified values into the object's hash.  The argument must
+be a reference to a hash, whose keys and values are copied into the
+object's hash.
+
+If the object was created with the I<value_optional> attribute enabled,
+keys may be inserted with no defined values.
+
+If the object was created with the I<value_type> attribute set to
+either "ARRAY" or "HASH", then the hash values must be references to
+the corresponding structure type.  If the values are lists, inserting a
+new list with the same key will append the new list to the existing list.
+If the values are hashes, inserting a new hash with the same key will
+insert the new key/value pairs into the existing value hash.
+
+If the object was created with the I<value_abspath> attribute enabled,
+the hash values must be absolute path strings.
+
+If the object was created with the I<key_abspath> attribute enabled, the
+hash keys must be absolute path strings.
+
+=item set()
+
+The same as insert(), except that the existing hash is emptied before
+inserting the new data.
+
+=item unset()
+
+Sets the object's value to an empty hash.
+
+=item delete()
+
+Deletes a specific hash key.  The argument can be a scalar or a
+reference to a list, in which case all of the keys in the list are
+deleted.
+
+=back
+
+=back
 
 =head1 AUTHOR
 
